//
//  GameScene.swift
//  TowerDefense
//
//  Created by 杉本裕樹 on 2015/10/03.
//  Copyright (c) 2015年 MyName. All rights reserved.
//

import SpriteKit
import Foundation
import GameplayKit
import ReplayKit

class GameScene: SKScene, SKPhysicsContactDelegate {
    var stage = 1
    var charData = [(factory: CharFactory, position: CGPoint)]()
    enum State {
        case playing
        case gameClear
        case gameOver
    }
    var state = State.playing
    var enemyList = EnemyList()
    var routes = [float2]()
    var chars = [Char]()
    
    override func didMove(to view: SKView) {
        physicsWorld.gravity = CGVector(dx: 0, dy: 0)
        physicsWorld.contactDelegate = self
        
        let fieldImageLength = view.frame.width / 10
        let field = FieldFactory().createField(view.frame.size, fieldImageLength: fieldImageLength)
        field.nodes.forEach {
            addChild($0)
        }
        
        routes = routesWithField(field)
        
        (0...10).forEach {
            perform(#selector(GameScene.createEnemy), with: nil, afterDelay: Double($0))
        }
    }
    
    func createChars(_ charData: [(factory: CharFactory, position: CGPoint)]) {
        charData.forEach {
            let char = $0.factory.createChar()
            char.position = $0.position
            char.physicsBody = SKPhysicsBody(rectangleOf: char.size)
            char.physicsBody?.contactTestBitMask = 0x1
            addChild(char)
            chars.append(char)
            self.charData.append($0)
        }
    }
    
    func createEnemy() {
        guard let view = view else {
            return
        }
        
        let enemy = enemyFactory().createEnemy()
        var routes = self.routes
        var prevPosition = routes.removeFirst()
        let actions = routes.map { p -> SKAction in
            let dx = p.x - prevPosition.x
            let dy = p.y - prevPosition.y
            let duration = Double(sqrt(dx * dx + dy * dy) / 100)
            prevPosition = p
            return SKAction.move(to: CGPoint(x: Double(p.x), y: Double(p.y)), duration: duration)
        }
        
        let fieldImageLength = view.frame.width / 10
        enemy.name = "enemy"
        enemy.position = CGPoint(x: fieldImageLength * 2, y: view.frame.height)
        enemy.physicsBody = SKPhysicsBody(rectangleOf: enemy.size)
        enemy.run(SKAction.sequence(actions)) {
            self.state = .gameOver
            
            let myLabel = SKLabelNode(fontNamed: "HiraginoSans-W6")
            myLabel.text = "ゲームオーバー"
            myLabel.fontSize = 45
            myLabel.position = CGPoint(x:self.frame.midX, y:self.frame.midY - 20)
            self.addChild(myLabel)
        }
        addChild(enemy)
        
        enemyList.appendEnemy(enemy)
    }
    
    func enemyFactory() -> EnemyFactory {
        switch stage {
        case 1: return Enemy1Factory()
        case 2: return Enemy2Factory()
        case 3: return Enemy3Factory()
        default: return EnemyFactory()
        }
    }
    
    override func update(_ currentTime: TimeInterval) {
        chars.forEach { char in
            char.stateMachine.update(withDeltaTime: currentTime)
            if let charState = char.stateMachine.currentState as? Char.CharState, charState.enableToAttack() {
                enemyList.enemiesCloseToPoint(char.frame.origin, distance: 50).forEach {
                    $0.life -= charState.power
                    char.stateMachine.enterState(Char.StayState.self)
                    if $0.life <= 0 {
                        $0.physicsBody?.node?.removeFromParent()
                        $0.physicsBody?.node?.removeAllActions()
                        
                        if enemyList.isAllEnemyRemoved() {
                            if stage < 3 {
                                let scene = SetCharScene()
                                scene.stage = self.stage + 1
                                scene.scaleMode = .resizeFill
                                scene.createChars(charData)
                                self.view?.presentScene(scene)
                                return
                            }
                            
                            state = .gameClear
                            
                            let myLabel = SKLabelNode(fontNamed: "HiraginoSans-W6")
                            myLabel.text = "ゲームクリア"
                            myLabel.fontSize = 45
                            myLabel.position = CGPoint(x:self.frame.midX, y:self.frame.midY - 20)
                            addChild(myLabel)
                            
                            let rootViewController = UIApplication.shared.delegate?.window??.rootViewController
                            RPScreenRecorder.shared().stopRecording { viewController, error in
                                viewController?.previewControllerDelegate = self
                                rootViewController?.present(viewController!, animated: true, completion: nil)
                            }
                        }
                    }
                }
            }
        }
    }
    
    private func routesWithField(_ field: Field) -> [vector_float2] {
        let fields = children.filter { $0.name == "Field0" }
        let obstacles = SKNode.obstacles(fromNodePhysicsBodies: fields)
        let graph = GKObstacleGraph(obstacles: obstacles, bufferRadius: 10)
        let start = GKGraphNode2D(point: vector_float2(Float(field.start.x), Float(field.start.y)))
        let end = GKGraphNode2D(point: vector_float2(Float(field.end.x), Float(field.end.y)))
        graph.connectNode(usingObstacles: start)
        graph.connectNode(usingObstacles: end)
        let nodes = graph.findPath(from: start, to: end)
        return nodes.flatMap { $0 as? GKGraphNode2D }.map { $0.position }
    }
}

extension GameScene: RPPreviewViewControllerDelegate {
    func previewControllerDidFinish(_ previewController: RPPreviewViewController) {
        DispatchQueue.main.async(execute: {
            previewController.dismiss(animated: true, completion: nil)
        })
    }
}
